


//#include "sdk_config.h"
#include "bluink_usb.h"
#include "app_usbd_hid_u2f_desc.h"
//#include "nrf_drv_power.h"
#include "boards.h"
#include "nrf_log.h"

int m_report_received = 0;
bool m_report_pending  = false;

static int m_totalLen = 0;

/**
 * @brief  USB connection status
 */
static bool m_usb_connected = false;

/**
 * @brief User event handler, HID keyboard
 */
static void hid_u2f_user_ev_handler(app_usbd_class_inst_t const * p_inst,
                                    app_usbd_hid_user_event_t event);


/**
 * @brief Reuse HID mouse report descriptor for HID generic class
 */
APP_USBD_HID_GENERIC_SUBCLASS_REPORT_DESC(hid_u2f_desc, APP_USBD_HID_U2F_REPORT_DSC());

ret_code_t u2f_hid_if_recv( uint32_t timeout );

static const app_usbd_hid_subclass_desc_t * reports[] = {&hid_u2f_desc};

#define MESSAGE_BUFFER_MAX  300
uint8_t g_messageBuffer[ MESSAGE_BUFFER_MAX ];
int g_messageSize = 0;

/**
 * @brief HID generic class endpoint number.
 * */
#define U2F_HID_EPIN       NRF_DRV_USBD_EPIN5
#define U2F_HID_EPOUT      NRF_DRV_USBD_EPOUT5

#define U2F_REPORT_OUT_MAXSIZE  63
#define HID_RPT_SIZE  64
/**
 * @brief HID generic class endpoints count.
 * */
#define HID_U2F_EP_COUNT  2

#define U2F_REPORT_IN_QUEUE_SIZE    1

/**
 * @brief List of HID generic class endpoints.
 * */
#define U2F_ENDPOINT_LIST()                                  \
(                                                        \
        U2F_HID_EPIN,                                    \
        U2F_HID_EPOUT                                    \
)

/**
 * @brief Global HID generic instance
 */

APP_USBD_HID_GENERIC_GLOBAL_DEF(m_app_hid_u2f,
                                APP_USBD_INTERFACE_U2F,
                                hid_u2f_user_ev_handler,
                                U2F_ENDPOINT_LIST(),
                                reports,
                                U2F_REPORT_IN_QUEUE_SIZE,
                                U2F_REPORT_OUT_MAXSIZE,
                                APP_USBD_HID_SUBCLASS_NONE,
                                APP_USBD_HID_PROTO_GENERIC);



const app_usbd_hid_generic_t *bluink_get_global_hid_u2f( void )
{
    return &m_app_hid_u2f;
}


/**
 * @brief Class specific event handler.
 *
 * @param p_inst    Class instance.
 * @param event     Class specific event.
 * */
static void hid_u2f_user_ev_handler(app_usbd_class_inst_t const * p_inst,
                                    app_usbd_hid_user_event_t event)
{
    switch (event)
    {
        case APP_USBD_HID_USER_EVT_OUT_REPORT_READY:
        {
            m_report_received += 1;
            size_t recv_size;            
            uint8_t *p_recv_buf = (uint8_t *)app_usbd_hid_generic_out_report_get(&m_app_hid_u2f, &recv_size);
            NRF_LOG_INFO("OUT_REPORT_READY #rec:%d, %02x %02x", m_report_received );
            if ( recv_size != HID_RPT_SIZE ) return;
            if ( g_messageSize + recv_size < MESSAGE_BUFFER_MAX ) {
                memcpy( g_messageBuffer + g_messageSize, p_recv_buf, recv_size );
                g_messageSize += recv_size;
            }
            break;
        }
        case APP_USBD_HID_USER_EVT_IN_REPORT_DONE:
        {
            NRF_LOG_INFO("IN_REPORT_DONE");
            m_report_pending = false;
            break;
        }
        case APP_USBD_HID_USER_EVT_SET_BOOT_PROTO:
        {
            UNUSED_RETURN_VALUE(hid_generic_clear_buffer(p_inst));
            NRF_LOG_INFO("SET_BOOT_PROTO");
            break;
        }
        case APP_USBD_HID_USER_EVT_SET_REPORT_PROTO:
        {
            UNUSED_RETURN_VALUE(hid_generic_clear_buffer(p_inst));
            NRF_LOG_INFO("SET_REPORT_PROTO");
            break;
        }
        default:
            break;
    }

}

/**
 * @brief USBD library specific event handler.
 *
 * @param event     USBD library event.
 * */
static void usbd_user_ev_handler(app_usbd_event_type_t event)
{
    switch (event)
    {
        case APP_USBD_EVT_DRV_SOF:
            break;
        case APP_USBD_EVT_DRV_RESET:
            m_report_pending = false;
            break;
        case APP_USBD_EVT_DRV_SUSPEND:
            m_report_pending = false;
            // Allow the library to put the peripheral into sleep mode
            app_usbd_suspend_req();
            bsp_board_leds_off();
            break;
        case APP_USBD_EVT_DRV_RESUME:
            m_report_pending = false;
            break;
        case APP_USBD_EVT_STARTED:
            m_report_pending = false;
            bsp_board_led_on(BSP_BOARD_LED_0);
            break;
        case APP_USBD_EVT_STOPPED:
            app_usbd_disable();
            bsp_board_leds_off();
            break;
        case APP_USBD_EVT_POWER_DETECTED:
            NRF_LOG_INFO("USB power detected");
            if (!nrf_drv_usbd_is_enabled())
            {
                app_usbd_enable();
            }
            break;
        case APP_USBD_EVT_POWER_REMOVED:
            NRF_LOG_INFO("USB power removed");
            app_usbd_stop();
            break;
        case APP_USBD_EVT_POWER_READY:
            NRF_LOG_INFO("USB ready");
            app_usbd_start();
            break;
        default:
            break;

    }
}

static const app_usbd_config_t usbd_config = {
    .ev_state_proc = usbd_user_ev_handler,
};



/**
 * @brief Auxiliary internal macro
 *
 * Macro used only in @ref init_bsp to simplify the configuration
 */
#define INIT_BSP_ASSIGN_RELEASE_ACTION(btn)                      \
    APP_ERROR_CHECK(                                             \
        bsp_event_to_button_action_assign(                       \
            btn,                                                 \
            BSP_BUTTON_ACTION_RELEASE,                           \
            (bsp_event_t)CONCAT_2(BSP_USER_EVENT_RELEASE_, btn)) \
    )


void bluink_usb_init( void )
{
    ret_code_t ret;

    ret = app_usbd_init(&usbd_config);
    APP_ERROR_CHECK(ret);

    app_usbd_class_inst_t const * class_inst_u2f;
    class_inst_u2f = app_usbd_hid_generic_class_inst_get(&m_app_hid_u2f);

    ret = app_usbd_class_append(class_inst_u2f);
    APP_ERROR_CHECK(ret);

}

void bluink_usb_start(void)
{
    //NRF_LOG_INFO("No USB power detection enabled\r\nStarting USB now");

    app_usbd_enable();
    app_usbd_start();
    m_usb_connected = true;
}


bool bluink_usb_connection_handle(bool last_usb_conn_status)
{
    if (last_usb_conn_status != m_usb_connected)
    {
        last_usb_conn_status = m_usb_connected;
        m_usb_connected ? app_usbd_start() : app_usbd_stop();
        NRF_LOG_INFO("USB library %s.", (uint32_t)(m_usb_connected ? "started" : "stopped"));
    }

    return last_usb_conn_status;
}


ret_code_t bluink_send_hid_u2f_frame( uint8_t *buffer, int length )
{

    const app_usbd_hid_generic_t *p_hid = bluink_get_global_hid_u2f();

    //app_usbd_hid_access_lock(p_u2f_ctx->hid_ctx);

    // Let's pad the rest of the report buffer with zeros
    //memset( p_u2f_ctx ->report_buff + length, 0, U2F_REPORT_SIZE - length );
    //INJ_circular_getBytes( cb, p_u2f_ctx ->report_buff, length );
    //p_u2f_ctx ->bufferChanged = 1;

    //app_usbd_hid_access_unlock(p_u2f_ctx->hid_ctx);

    ret_code_t ret = app_usbd_hid_generic_in_report_set(p_hid, buffer, length);

    NRF_LOG_INFO("U2F Transfer IN: %d bytes, status: %d", length, ret );
    if (ret == NRF_SUCCESS)
    {
        m_report_pending = true;
    } else {
        NRF_LOG_INFO("IN report status: %d", ret );
    }

    return ret;
}



void bluink_process_usb_events( void )
{

//    hid_retry_handler();
    while ( 1 ) {
        if ( !app_usbd_event_queue_process() ) {
             break;
        }
    }
    if ( m_report_received > 0 ) {
        // See if we have enough data to process.
        m_report_received = 0;
        int messageSize = g_messageSize;
        if ( messageSize > 6 ) {
            int payloadSize = (g_messageBuffer[ 5 ] << 8) | g_messageBuffer[ 6 ];
            if ( payloadSize < messageSize ) {
                NRF_LOG_INFO("Message recevied");
                // We just echo it back!
    
                for ( int offset = 0; offset < messageSize; offset += HID_RPT_SIZE ) {
                    while ( m_report_pending ) {
                        app_usbd_event_queue_process();
                    }
                    bluink_send_hid_u2f_frame( g_messageBuffer + offset, HID_RPT_SIZE );
                }
                g_messageSize = 0;
            }
        } else {
            NRF_LOG_INFO("Message size unexpected: %d", g_messageSize );
        }
    }
}



