/**
 * Copyright (c) 2016 - 2017, Nordic Semiconductor ASA
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 * 
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 * 
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#ifndef APP_USBD_HID_U2F_DESC_H__
#define APP_USBD_HID_U2F_DESC_H__

#include "app_usbd_descriptor.h"

#ifdef __cplusplus
extern "C" {
#endif


/**
 * @brief Initializer of interface descriptor for HID keyboard class.
 *
 * @param interface_number  Interface number.
 */
#define APP_USBD_HID_U2F_INTERFACE_DSC(interface_number)            \
        APP_USBD_HID_INTERFACE_DSC(interface_number,                \
                                   2,                               \
                                   0,      \
                                   0)

/**
 * @brief Initializer of HID descriptor for HID keyboard class.
 *
 * @param ...   Report descriptor item.
 */
#define APP_USBD_HID_U2F_HID_DSC(...)       \
        APP_USBD_HID_HID_DSC(__VA_ARGS__)
/**
 * @brief Initializer of endpoint descriptor for HID keyboard class.
 *
 * @param endpoint  Endpoint number.
 */
#define APP_USBD_HID_U2F_EP_DSC(endpoint)   \
        APP_USBD_HID_EP_DSC(endpoint, 64, 1)


/**
 * @brief U2F Report Descriptor
 *
 */
#ifdef BLUINK_POS
#define APP_USBD_HID_U2F_REPORT_DSC() {   \
	        0x06, 0xb0, 0x1c,	/* USAGE_PAGE (LCBO proof of concept) */\
	        0x09, 0x01, 		/* USAGE (U2F HID Authenticator Device) */\
	        0xa1, 0x01, 		/* COLLECTION (Application) */\
	        0x09, 0x20, 		/*   USAGE (Input Report Data) */\
	        0x15, 0x00, 		/*   LOGICAL_MINIMUM (0) */\
	        0x26, 0xff, 0x00, 	/*   LOGICAL_MAXIMUM (255) */\
	        0x75, 0x08, 		/*   REPORT_SIZE (8) */\
	        0x95, 0x40, 		/*   REPORT_COUNT (64) */\
	        0x81, 0x02, 		/*   INPUT (Data,Var,Abs) */\
	        0x09, 0x21, 		/*   USAGE (Output Report Data) */\
	        0x15, 0x00, 		/*   LOGICAL_MINIMUM (0) */\
	        0x26, 0xff, 0x00, 	/*   LOGICAL_MAXIMUM (255) */\
	        0x75, 0x08, 		/*   REPORT_SIZE (8) */\
	        0x95, 0x40, 		/*   REPORT_COUNT (64) */\
	        0x91, 0x02, 		/*   OUTPUT (Data,Var,Abs) */\
	        0xc0 			/* END_COLLECTION */\
}
#else
#define APP_USBD_HID_U2F_REPORT_DSC() {   \
	        0x06, 0xd0, 0xf1,  	/* USAGE_PAGE (FIDO) */\
	        0x09, 0x01, 		/* USAGE (U2F HID Authenticator Device) */\
	        0xa1, 0x01, 		/* COLLECTION (Application) */\
	        0x09, 0x20, 		/*   USAGE (Input Report Data) */\
	        0x15, 0x00, 		/*   LOGICAL_MINIMUM (0) */\
	        0x26, 0xff, 0x00, 	/*   LOGICAL_MAXIMUM (255) */\
	        0x75, 0x08, 		/*   REPORT_SIZE (8) */\
	        0x95, 0x40, 		/*   REPORT_COUNT (64) */\
	        0x81, 0x02, 		/*   INPUT (Data,Var,Abs) */\
	        0x09, 0x21, 		/*   USAGE (Output Report Data) */\
	        0x15, 0x00, 		/*   LOGICAL_MINIMUM (0) */\
	        0x26, 0xff, 0x00, 	/*   LOGICAL_MAXIMUM (255) */\
	        0x75, 0x08, 		/*   REPORT_SIZE (8) */\
	        0x95, 0x40, 		/*   REPORT_COUNT (64) */\
	        0x91, 0x02, 		/*   OUTPUT (Data,Var,Abs) */\
	        0xc0 			/* END_COLLECTION */\
}
#endif

/** @} */

#ifdef __cplusplus
}
#endif

#endif /* APP_USBD_HID_U2F_DESC_H__ */
